from io import BytesIO
from typing import Optional


class Image(object):

    """
    An AsyncDagpi Image Returned
    This has special properties that can enhance the experience.

    .. container:: Attributes

        **Attributes**

        image: :class:`io.BytesIO`
            BytesIO object with Image
        format: :class:`str`
            String containing Image Format
            Either PNG or GIF
        process_time: :class:`str`
            String for Time Taken by API to process Request
        original_image: :class:`str`
            URL for the image passed to the API
    """

    def __init__(self, byt: bytes, image_format: str, process_time: str,
                 original_url: str):
        """
        Initialise an asyncdagpi.Image
        """
        self.image: BytesIO = BytesIO(byt)
        self.format: str = image_format
        self.process_time: str = process_time
        self.original_image: str = original_url

    def size(self) -> int:
        """
        Returns the Size of the Image
        """
        return self.image.__sizeof__()

    def __repr__(self) -> str:
        """
        A description of the Image
        """
        return f"<asyncdagpi.Image format={self.format}>"

    def read(self) -> bytes:
        """
        Get raw Image bytes
        :return: :class:`bytes`
        """
        return self.image.read()

    def write(self, name: str, *, path: Optional[str] = None) -> None:
        """
        Writes the Image Object to file

        :param name: :class:`str`
            the name of the file to save
        kwargs: Optional To Pass in
            - path: path to write to can be relative/absolute. default is ./
        """
        path = path or "."
        with open("{}/{}.{}".format(path, name, self.format), 'wb') as file:
            file.write(self.read())



class AsyncDagpiException(Exception):
    """
    AsyncDagpi base exception class use this base class to catch any AsyncDagpi errors.
    """
    pass


class AsyncDagpiHttpException(AsyncDagpiException):
    """
    AsyncDagpi base exception class use this base class to catch any AsyncDagpi errors.
    """

    def __init__(self, status: int, message: str) -> None:
        self.status: int = status
        self.message: str = message
        super().__init__(message)


class ApiError(AsyncDagpiHttpException):
    """
    Raised when Dagpi has an error it does not know how to handle
    """

    def __init__(self, message: str) -> None:
        super().__init__(500, message)


class BadUrl(AsyncDagpiHttpException):
    """
    Exception raised when the URL is poorly framed or not of type String
    """

    def __init__(self, message: str) -> None:
        super().__init__(400, message)


class Unauthorised(AsyncDagpiHttpException):
    """
    Raised for an API 401
    """

    def __init__(self, message: str) -> None:
        super().__init__(401, message)


class RateLimited(AsyncDagpiHttpException):
    """
    You are exceeding the API's rate limits and built in Ratelimit handler
    Essentially a 429
    """

    def __init__(self, message: str) -> None:
        super().__init__(429, message)


class FileTooLarge(AsyncDagpiHttpException):
    """
    The image at source exceeds 10 MegaBytes
    """

    def __init__(self, message: str) -> None:
        super().__init__(413, message)


class InvalidFeature(AsyncDagpiException):
    """
    The feature chosen is not valid
    """
    pass


class ImageUnaccesible(AsyncDagpiHttpException):
    """
    The API was unable to get an image at your location
    """

    def __init__(self, error_code: int, message: str):
        """
        Initialise the ImageUnaccessible Error
        """
        self.error_code: int = error_code
        self.message: str = message
        super(ImageUnaccesible, self).__init__(self.error_code,
                                               f'{self.error_code} ---> \
        {self.message}')

    def __str__(self) -> str:
        """
        String Explaining the error
        :return: :class:`str`
        """
        return f'{self.error_code} ---> {self.message}'


class ParameterError(AsyncDagpiHttpException):
    """
    Parameters passed were not Sufficient
    """

    def __init__(self, message: str) -> None:
        super().__init__(400, message)



# -*- coding: utf-8 -*-

"""

Asyncdagpi
~~~~~~~~~~~~~~~~~~~
An async wrapper for dagpi
:copyright: (c) 2020 Daggy
:license: MIT, see LICENSE for more details.
"""

import logging
from typing import NamedTuple
from .image_features import ImageFeatures
from .client import Client
from .image import Image
from .objects import WTP, Logo, PickupLine, BaseDagpiObject, Ratelimits, Headline, Captcha, Typeracer

__title__: str = "asyncdagpi"
__author__: str = "Daggy1234"
__license__: str = "MIT"
__copyright__: str = "Copyright 2021 Daggy1234"
__version__: str = '4.1.5'


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


version_info: VersionInfo = VersionInfo(major=4, minor=1, micro=5, releaselevel='alpha',
                                        serial=0)

logging.getLogger(__name__).addHandler(logging.NullHandler())



from asyncdagpi.objects import Ratelimits
import asyncio
from asyncio.events import AbstractEventLoop
import logging
from typing import Dict, Optional, Any

import aiohttp

from aiohttp import __version__ as aiov
from aiohttp.client import ClientSession

from . import errors
import sys
from .image import Image

error_dict: Dict[int, errors.AsyncDagpiHttpException] = {
    400: errors.ParameterError("Parameters passed were incorrect"),
    413: errors.FileTooLarge("The Image Passed is too large"),
    500: errors.ApiError("Internal Server Error"),
    429: errors.RateLimited("You are being Rate_limited"),
    403: errors.Unauthorised("403 Returned")
}


log: logging.Logger = logging.getLogger(__name__)


class HTTP:

    """
    HTTP Client
    -----------
    Represents an HTTP client sending HTTP requests to the dagpi API.
        .. _aiohttp session:
https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session
        Parameters
        ----------
        :param token: :class:`str`
            A dagpi Token from https://dagpi.xyz
        *args:
            **session : Optional[aiohttp session]
                The session used to request to the API
            **loop: Optional[asyncio loop]
                The asyncio loop to use
    """

    __slots__ = ("client", "base_url", "token", "loop",
                 "user_agent", "logging", "ratelimits")

    def __init__(
        self,
        token: str,
        session: Optional[aiohttp.ClientSession],
        loop: Optional[asyncio.AbstractEventLoop]
    ):
        """
        Initialise the dagpi http client
        """
        self.base_url: str = "https://api.dagpi.xyz"
        self.token: str = token
        self.loop: AbstractEventLoop = loop or asyncio.get_event_loop()
        self.client: ClientSession = session or aiohttp.ClientSession(loop=loop)
        self.ratelimits: Ratelimits = Ratelimits(None, None, None)
        from asyncdagpi import __version__
        self.user_agent: str = f"AsyncDagpi v{__version__} Python/{sys.version_info[0]}.{sys.version_info[1]} \
            aiohttp/{aiov}"

    async def data_request(self, url: str, *, image: Optional[bool] = None) -> Dict[str, Any]:
        """

        url: :class:`str`
            url to request
        :return: :class:`Dict`
            Python Dictionary
        """

        if not self.token:
            raise errors.Unauthorised("Please Provide a dagpi token")

        headers = {
            "Authorization": self.token,
            'User-Agent': self.user_agent
        }

        request_url = self.base_url + "/data/" + url
        if image:
            request_url = self.base_url + "/image/"
        async with self.client.get(request_url, headers=headers) as resp:
            self.ratelimits = Ratelimits.from_dict(resp.headers)
            if 300 >= resp.status >= 200:
                if resp.headers["Content-Type"] != "application/json":
                    raise errors.ApiError(f"{resp.status}. \
                    Request was great but Dagpi did not send a JSON")
                js: Dict[str, Any] = await resp.json()
                return js

            else:
                try:
                    error = error_dict[resp.status]
                    raise error
                except KeyError:
                    raise errors.ApiError("Unknown API Error Occurred")

    async def image_request(self, url: str, params: Dict[str, str]) -> Image:
        """

        url: :class:`str`
            A string containing the URL
        params: :class:`Dict`
            A dictionary of the URL parameters
        :return: :class:`asyncdagpi.Image`
            Asyncdagpi Image Object
        """

        if not self.token:
            raise errors.Unauthorised("Please Provide a dagpi token")

        headers = {
            "Authorization": self.token,
            'User-Agent': self.user_agent
        }

        request_url = self.base_url + "/image" + url
        async with self.client.get(request_url, headers=headers,
                                   params=params) as resp:
            self.ratelimits = Ratelimits.from_dict(resp.headers)
            if 300 >= resp.status >= 200:
                if resp.headers["Content-Type"].lower() not in [
                    "image/png",
                    "image/gif",
                ]:
                    raise errors.ApiError(f"{resp.status}. \
                Request was great but Dagpi did not send an Image back")
                form: str = resp.headers["Content-Type"].replace("image/",
                                                                 "")
                resp_time = resp.headers["X-Process-Time"][:5]
                raw_byte = await resp.read()
                log.info(
                    f'[Dagpi Image] GET {resp.url} has returned {resp.status}')
                return Image(raw_byte, form, resp_time,
                             params["url"])

            else:
                try:
                    error = error_dict[resp.status]
                    raise error
                except KeyError:
                    js = await resp.json()
                    if resp.status == 415:
                        raise errors.ImageUnaccesible(415, js["message"])
                    elif resp.status == 400:
                        raise errors.ParameterError(js["message"])
                    elif resp.status == 422:
                        try:
                            mstr = ""
                            for val in js["detail"]:
                                base = f'{val["loc"][1]} is {val["type"]}'
                                mstr += (base + "\t")
                            raise errors.ParameterError(mstr)
                        except KeyError:
                            raise errors.ApiError(
                                "API was unable to manipulate the Image")
                    else:
                        raise errors.ApiError("Unknown API Error Occurred")
                finally:
                    log.error(f'[Dagpi Image] GET {resp.url} has returned {resp.status}')

    async def close(self) -> None:
        await self.client.close()



from typing import Type, TypeVar, Optional

IF = TypeVar('IF', bound='ImageFeatures')


class ImageFeatures:
    """
    Base ImageFatures that asyncdagpi has
    """
    def __init__(self, value: str, description: str):
        """
        Initialise the dagpi Image feature
        """
        self.value: str = value
        self.description: str = description

    def __str__(self) -> str:
        """
        Get a string of the feature
        :return: :class:`string`
        """
        return self.value.replace("/", "")

    def __repr__(self) -> str:
        """
        get a string describing the object
        :return: :class:`str`
        """
        return f"<asyncdagpi.ImageFeature feature={self.value.replace('/', '')} >"

    @classmethod
    def from_path(cls: Type[IF], path: str) -> IF:
        """
        Construct an Image Feature from it's path.
        """
        return cls(path, "")

    @classmethod
    def mirror(cls: Type[IF]) -> IF:
        """
        Flip image horizontally (left to right).
        """
        return cls("/mirror/", "Flip image horizontally (left to right).")

    @classmethod
    def flip(cls: Type[IF]) -> IF:
        """
        Flip the image vertically (top to bottom).
        """
        return cls("/flip/", "Flip the image vertically (top to bottom).")

    @classmethod
    def pixel(cls: Type[IF]) -> IF:
        """
        Pixelate an Image
        """
        return cls("/pixel/", "Pixelate an Image")

    @classmethod
    def colors(cls: Type[IF]) -> IF:
        """
        Analyse and get an Image's Color Top 5 Colors
        """
        return cls("/colors/", "Analyse and get an Image's Color Top 5 Colors")

    @classmethod
    def wanted(cls: Type[IF]) -> IF:
        """
        Get a wanted poster of an Image.
        """
        return cls("/wanted/", "Get a wanted poster of an Image")

    @classmethod
    def triggered(cls: Type[IF]) -> IF:
        """
        Get a triggered Image gif
        """
        return cls("/triggered/", "Get a triggered Image gif")

    @classmethod
    def america(cls: Type[IF]) -> IF:
        """
        The waving american flag on an image.gif
        """
        return cls("/america/", "The waving american flag on an image.gif")

    @classmethod
    def communism(cls: Type[IF]) -> IF:
        """
        Glory of the soviet Union on an image.gif
        """
        return cls("/communism/", "Glory of the soviet Union on an image.gif")

    @classmethod
    def bomb(cls: Type[IF]) -> IF:
        """
        Kaboom
        """
        return cls("/bomb/", "Kaboom")

    @classmethod
    def wasted(cls: Type[IF]) -> IF:
        """
        GTA V Wasted screen.
        """
        return cls("/wasted/", "GTA V Wasted screen")

    @classmethod
    def five_guys_one_girl(cls: Type[IF]) -> IF:
        """
        No I have never heard of this meme. Takes in 2 Image URL.
        Needs:
            - url
            - url2
        """
        return cls("/5g1g/",
                   "No I have never heard of this meme. Takes in 2 Image URL")

    @classmethod
    def slap(cls: Type[IF]) -> IF:
        """
        Have someone slap someone.
        Needs:
            - url
            - url2
        """
        return cls("/slap/", "Have someone slap someone.")

    @classmethod
    def why_are_you_gay(cls: Type[IF]) -> IF:
        """
        The infamous Why are you gay. Takes in 2 Image URL.
        Needs:
            - url
            - url2
        """
        return cls("/whyareyougay/",
                   "The infamous  Why are you gay. Takes in 2 Image URL")

    @classmethod
    def invert(cls: Type[IF]) -> IF:
        """
        Invert an Image.
        """
        return cls("/invert/", "Invert an Image")

    @classmethod
    def sobel(cls: Type[IF]) -> IF:
        """
        Cool SOBEL filter for Images. Only png's.
        """
        return cls("/sobel/", "Cool SOBEL filter for Images.")

    @classmethod
    def triangle(cls: Type[IF]) -> IF:
        """
        Cool triangle edge detection for Images. Only png's.
        """
        return cls("/triangle/", "Cool triangle edge detection for Images.")

    @classmethod
    def hog(cls: Type[IF]) -> IF:
        """
        Histogram of Oriented Gradients.
        """
        return cls("/hog/", "Histogram of Oriented Gradients")

    @classmethod
    def blur(cls: Type[IF]) -> IF:
        """
        Blurs an entire Image.
        """
        return cls("/blur/", "Blurs an entire Image")

    @classmethod
    def rgb(cls: Type[IF]) -> IF:
        """
        Get RGB information from an image.
        """
        return cls("/rgb/", "Get RGB information from an image")

    @classmethod
    def angel(cls: Type[IF]) -> IF:
        """
        Make an Image Angelic.
        """
        return cls("/angel/", "Make an Image angelic")

    @classmethod
    def satan(cls: Type[IF]) -> IF:
        """
        Make an Image the devil.
        """
        return cls("/satan/", "Make an Image the devil")

    @classmethod
    def hitler(cls: Type[IF]) -> IF:
        """
        Make an Image as bad as hitler.
        """
        return cls("/hitler/", "Make an Image as bad as hitler")

    @classmethod
    def obama(cls: Type[IF]) -> IF:
        """
        The obama meme of someone awarding themselves.
        """
        return cls("/obama/", "The obama meme of someone awarding themselves")

    @classmethod
    def bad(cls: Type[IF]) -> IF:
        """
        This image is bad.
        """
        return cls("/bad/", "This image is bad")

    @classmethod
    def sith(cls: Type[IF]) -> IF:
        """
        Laughs in Sithlord
        """
        return cls("/sith/", "Laughs in Sithlord")

    @classmethod
    def jail(cls: Type[IF]) -> IF:
        """
        Put an Image behind bars.
        """
        return cls("/jail/", "Put an Image behind bars")

    @classmethod
    def fedora(cls: Type[IF]) -> IF:
        """
        Tips. Fedora
        """
        return cls("/fedora/", "Tips fedora")

    @classmethod
    def delete(cls: Type[IF]) -> IF:
        """
        Delete some trash
        """
        return cls("/delete/", "Delete some trash")

    @classmethod
    def shatter(cls: Type[IF]) -> IF:
        """
        Broken like glass
        """
        return cls("/shatter/", "Broken like glass")

    @classmethod
    def gay(cls: Type[IF]) -> IF:
        """
        represent! Pastes a pride flag on an image.
        """
        return cls("/gay/", "represent! Pastes a pride flag on an image")

    @classmethod
    def pride(cls: Type[IF]) -> IF:
        """
        Pride flag overlay
        Needs:
            - flag (one of these flags):
                * asexual

                * bisexual

                * gay

                * genderfluid

                * genderqueer

                * intersex

                * lesbian

                * nonbinary

                * progress

                * pan

                * trans

                * agender

                * ally

                * polysexual
        """
        return cls("/pride/", "Pride flag overlay")

    @classmethod
    def trash(cls: Type[IF]) -> IF:
        """
        Makes an Image trash.
        """
        return cls("/trash/", "Makes an Image trash")

    @classmethod
    def deepfry(cls: Type[IF]) -> IF:
        """
        Deepfries an Image.
        """
        return cls("/deepfry/", "Deepfries an Image")

    @classmethod
    def ascii(cls: Type[IF]) -> IF:
        """
        Turns an Image into a ascii-fied image.
        """
        return cls("/ascii/", "Turns an Image into a ascii-fied image")

    @classmethod
    def charcoal(cls: Type[IF]) -> IF:
        """
        Turns an Image into a charcoal sketch.
        """
        return cls("/charcoal/", "Turns an Image into a charcoal sketch")

    @classmethod
    def poster(cls: Type[IF]) -> IF:
        """
        Posterizes an image.
        """
        return cls("/poster/", "Posterizes an image")

    @classmethod
    def sepia(cls: Type[IF]) -> IF:
        """
        Makes an image sepia tone.
        """
        return cls("/sepia/", "Makes an image sepia tone")

    @classmethod
    def polaroid(cls: Type[IF]) -> IF:
        """
        Frames an Image like a printed polaroid.
        """
        return cls("/polaroid/", "Frames an Image like a printed polaroid")

    @classmethod
    def glitch(cls: Type[IF]) -> IF:
        """
        *Cool Glitched image*
        """
        return cls("/glitch/", "*Cool Glitched image*")

    @classmethod
    def swirl(cls: Type[IF]) -> IF:
        """
        Swirls the center of Image.
        """
        return cls("/swirl/", "Swirls the center of Image")

    @classmethod
    def paint(cls: Type[IF]) -> IF:
        """
        Turns an Image into an oil painting.
        """
        return cls("/paint/", "Turns an Image into an oil painting")

    @classmethod
    def sketch(cls: Type[IF]) -> IF:
        """
        How an artist would create an image.
        """
        return cls("/sketch/", "How an artist would create an image.")

    @classmethod
    def spin(cls: Type[IF]) -> IF:
        """
        the gif goes round and round.
        """
        return cls("/spin/", "the gif goes round and round")

    @classmethod
    def dissolve(cls: Type[IF]) -> IF:
        """
        Thanos snapped and back.
        """
        return cls("/dissolve/", "Thanos snapped and back.")

    @classmethod
    def neon(cls: Type[IF]) -> IF:
        """
        A cool multicolored glowing neon sign from an image.
        """
        return cls("/neon/",
                   "A cool multicolored glowing neon sign form an image.")

    @classmethod
    def petpet(cls: Type[IF]) -> IF:
        """
        Pet Pet Pet.
        """
        return cls("/petpet/", "Pet Pet Pet.")

    @classmethod
    def comic(cls: Type[IF]) -> IF:
        """
        classic black and white comic.
        """
        return cls("/comic/", "Classic black and white commic.")

    @classmethod
    def burn(cls: Type[IF]) -> IF:
        """
        Burn an image untill there's molten remains.
        """
        return cls("/burn/", "Burn an image untill there's molten remains.")

    @classmethod
    def freeze(cls: Type[IF]) -> IF:
        """
        Get blasted by Mr.Freeze from batman.
        """
        return cls("/freeze/", "Get blasted by Mr.Freeze from batman.")

    @classmethod
    def earth(cls: Type[IF]) -> IF:
        """
        Become one with Rock.
        """
        return cls("/earth/", "Become one with rock.")

    @classmethod
    def night(cls: Type[IF]) -> IF:
        """
        Turns day into night on an Image.
        """
        return cls("/night/", "Turns day into night on an Image")

    @classmethod
    def magik(cls: Type[IF]) -> IF:
        """
        Magik an Image
        """
        return cls("/magik/", "Magik an Image")

    @classmethod
    def stringify(cls: Type[IF]) -> IF:
        """
        Stringify an Image
        """
        return cls("/stringify/", "Stringify an Image")

    @classmethod
    def rainbow(cls: Type[IF]) -> IF:
        """
        Rainbow light effect on image
        """
        return cls("/rainbow/", "Rainbow light effect on image")

    @classmethod
    def solar(cls: Type[IF]) -> IF:
        """
        Solarizes an Image.
        """
        return cls("/solar/", "Solarizes an Image")

    @classmethod
    def bonk(cls: Type[IF]) -> IF:
        """
        Bonk Someone
        """
        return cls("/bonk/", "Bonk Someone")

    @classmethod
    def captcha(cls: Type[IF]) -> IF:
        """
        Creates a realistic Captcha
        Needs:
            - text
            - image
        """
        return cls("/captcha/", "Creates a realistic Captcha")

    @classmethod
    def thought_image(cls: Type[IF]) -> IF:
        """
        Wraps Image and Text into a thought
        Needs:
            - text
            - image
        """
        return cls("/thoughtimage/", "Wraps Image and Text into a thought")

    @classmethod
    def tweet(cls: Type[IF]) -> IF:
        """
        Generates an accurate fake tweet
        Needs:
            - text
            - image
            - username
        """
        return cls("/tweet/", "Generates an accurate fake tweet")

    @classmethod
    def discord(cls: Type[IF]) -> IF:
        """
        Generated a realistic Discord Message
        Needs:
            - text
            - image
            - username
            - dark (boolean) default is true
        """
        return cls("/discord/", "Generated a realistic Discord Message")

    @classmethod
    def youtube(cls: Type[IF]) -> IF:
        """
        Generated a realistic Yutube comment
        Needs:
            - text
            - image
            - username
            - dark (boolean) default is true
        """
        return cls("/yt/", "Generated a realistic Yutube comment")

    @classmethod
    def retro_meme(cls: Type[IF]) -> IF:
        """
        old 2012 advice animals style meme
        Needs:
            - top_text
            - bottom_text
            - image
        """
        return cls("/retromeme/", "Generated a realistic Discord Message")

    @classmethod
    def motiv(cls: Type[IF]) -> IF:
        """
        old 2012 style Motiv memes
        Needs:
            - top_text
            - bottom_text
            - image
        """
        return cls("/motiv/", "old 2012 style Motiv memes")

    @classmethod
    def modern_meme(cls: Type[IF]) -> IF:
        """
        Mordern meme with text scaling
        Needs:
            - text
            - image
        """
        return cls("/modernmeme/", "Modern New meme style meme")

    @classmethod
    def mosiac(cls: Type[IF]) -> IF:
        """
        Turn an image into a roman mosiac
        Needs:
            - image
            - pixels
        """
        return cls("/mosiac/", "Turn an image into a roman mosiac")

    @classmethod
    def cube(cls: Type[IF]) -> IF:
        """
        Turn an image into a cube
        Needs:
            - image
        """
        return cls("/cube/", "Turn an image into a cube")

    @classmethod
    def lego(cls: Type[IF]) -> IF:
        """
        image out of lego bricks
        Needs:
            - image
        """
        return cls("/lego/", "image out of lego bricks")

    @classmethod
    def expand(cls: Type[IF]) -> IF:
        """
        blown out of proportions
        Needs:
            - image
        """
        return cls("/expand/", "blown out of proportions")

    @classmethod
    def elmo(cls: Type[IF]) -> IF:
        """
        burning elmo gif
        Needs:
            - image
        """
        return cls("/elmo/", "burning elmo gif")

    @classmethod
    def tv(cls: Type[IF]) -> IF:
        """
        burning elmo gif
        Needs:
            - image
        """
        return cls("/elmo/", "burning elmo gif")

    @classmethod
    def glitch_static(cls: Type[IF]) -> IF:
        """
        simpler static glitch
        Needs:
            - image
        """
        return cls("/elmo/", "simpler static glitch")

    @classmethod
    def album(cls: Type[IF]) -> IF:
        """
        turn image into album cover
        Needs:
            - image
        """
        return cls("/album/", "turn image into album cover")

    @classmethod
    def rain(cls: Type[IF]) -> IF:
        """
        for rainy days
        Needs:
            - image
        """
        return cls("/rain/", "for rainy days")

    @classmethod
    def from_string(cls: Type[IF], feature: str) -> Optional[IF]:
        """Get an image feature from a string

        Args:
            cls (Type[IF]): class stuff
            feature (str): string feature to try

        Returns:
            IF: Image Feature
        """
        try:
            image_feature = getattr(cls, feature)
            called = image_feature()
            return cls(called.value, called.description)
        except BaseException:
            return None



import re
import time
from typing import Dict, Optional, Any

from .errors import BadUrl, InvalidFeature
from .http import HTTP
from .image import Image
from .image_features import ImageFeatures
from .objects import Captcha, Typeracer, WTP, PickupLine, Logo, Headline, Ratelimits
from aiohttp import ClientSession
from asyncio import AbstractEventLoop


url_regex: str = r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]| " \
    r"(?:%[0-9a-fA-F][0-9a-fA-F]))+"


class Client:
    """Client for interacting with asyncdagpi.

    Attributes
    -----------

    token: :class:`str`
        Your Dagpi Api token

    session: Optional[:class:`aiohttp.ClientSession`]
        an aiohttp ClientSession for dagpi to use

    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        an asyncio event loop for the asyncdagpi to use

    .. note::
        This will initialise an AsyncDagpiClient that you can use for making requests.
        It is recommended to re-use the Client.

    ..note::
        The :class:`aiohttp.ClientSession` passed via constructor will have headers and authentication set.
        Do not pass one you plan to re-use for other things, lest you leak your login data.
    """

    def __init__(
        self,
        token: str,
        *,
        session: Optional[ClientSession] = None,
        loop: Optional[AbstractEventLoop] = None
    ):

        self.token: str = token
        self.session: Optional[ClientSession] = session
        self.loop: Optional[AbstractEventLoop] = loop
        self.http: HTTP = HTTP(self.token, loop=self.loop,
                               session=self.session)

    @staticmethod
    def url_test(url: str) -> None:
        if not isinstance(url, str):
            raise BadUrl("URL is not a String")

        regex = re.compile(
            url_regex,
            re.IGNORECASE)
        match = re.match(regex, url)
        if not match:
            raise BadUrl("URL did not pass Regex")

    @property
    def ratelimits(self) -> Ratelimits:
        """Get ratelimits for your client
        """
        return self.http.ratelimits

    async def image_process(
        self,
        feature: ImageFeatures,
        url: str,
        **kwargs: Any) \
            -> Image:
        """|coro|

        Parameters
        -----------
        feature: :class:`ImageFeature`
            a dagpi ImageFeature class

        url: :class:`str`
            the Url for the Image Passed

        kwargs:
            based on the Docs for your Feature chose the right
            extra kwargs like `text` or `username`

        Returns
        --------
        :class:`asyncdagpi.Image`
            Asyncdagpi Image Object
        """
        if not isinstance(feature, ImageFeatures):
            raise InvalidFeature(f"{feature} does not exist")
        self.url_test(url)
        dark = kwargs.get("dark")
        if dark is not None:
            kwargs["dark"] = str(dark)
        params: Dict[str, str] = {"url": url, **kwargs}
        return await self.http.image_request(feature.value, params=params)

    async def special_image_process(self, url: str) \
            -> Image:
        """|coro|

        Parameters
        -----------
        url: :class:`str`
            the Url for the Image Passed

        Returns
        --------
        :class:`asyncdagpi.Image`
            Asyncdagpi Image Object
        """
        self.url_test(url)
        params = {"url": url}
        return await self.http.image_request("/special/", params=params)

    async def wtp(self) -> WTP:
        """
        get a WTP data object
        :returns: :class:`asyncdagpi.WTP`
        """
        raw_data = await self.http.data_request("wtp")
        return WTP(raw_data)

    async def logo(self) -> Logo:
        """|coro|
        get a Logo data object

        Returns
        --------
        :class:`asyncdagpi.Logo`
            logo object
        """
        raw_data = await self.http.data_request("logo")
        return Logo(raw_data)

    async def roast(self) -> str:
        """|coro|
        Returns a string with a Roast

        Returns
        --------
        :class:`str`
        """
        obj = await self.http.data_request("roast")
        return obj["roast"]

    async def yomama(self) -> str:
        """|coro|
        Returns a YoMama Joke String

        Returns
        --------
        :class:`str`
        """
        obj = await self.http.data_request("yomama")
        return obj["description"]

    async def joke(self) -> str:
        """|coro|
        Gets a Funny Joke

        Returns
        --------
        :class:`str`
        """
        obj = await self.http.data_request("joke")
        return obj["joke"]

    async def fact(self) -> str:
        """|coro|
        Gets a Fun fact

        Returns
        --------
        :class:`str`
        """
        obj = await self.http.data_request("fact")
        return obj["fact"]

    async def eight_ball(self) -> str:
        """|coro|
        Gets an 8ball response

        Returns
        --------
        :class:`str`
        """
        obj = await self.http.data_request("8ball")
        return obj["response"]

    async def pickup_line(self) -> PickupLine:
        """|coro|
        Get a PickupLine

        Returns
        --------
        :class:`asyncdagpi.objects.PickupLine`
        """

        return PickupLine(await self.http.data_request("pickupline"))

    async def headline(self) -> Headline:
        """|coro|
        Get a Headline

        Returns
        --------
        :class:`asyncdagpi.objects.Headline`
        """

        return Headline(await self.http.data_request("headline"))

    async def captcha(self) -> Captcha:
        """|coro|

        Get a captcha

        Returns
        --------
        :class:`asyncdagpi.objects.Captcha`
        """

        return Captcha(await self.http.data_request("captcha"))

    async def typeracer(self) -> Typeracer:
        """|coro|
        Get a sentence on an image

        Returns
        --------
        :class:`asyncdagpi.objects.Typeracer`
        """

        return Typeracer(await self.http.data_request("typeracer"))

    async def waifu(self) -> Dict[str, Any]:
        """|coro|
        Get a Random Anime Waifu.
        Does not return a model due to sheer complexity and impracticality.

        Returns
        --------
        :class:`asyncdagpi.objects.PickupLine`
        """
        return await self.http.data_request("waifu")

    async def data_ping(self) -> float:
        """|coro|
        Returns a float with the Data API's ping

        Returns
        --------
        :class:`float`
            ping time for data api
        """
        start = time.perf_counter()
        await self.http.data_request("wtp")
        end = time.perf_counter()
        return (end - start)

    async def image_ping(self) -> float:
        """|coro|
        Returns a float with the Image API's ping

        Returns
        --------
        :class:`float`
            ping time for image api
        """
        start = time.perf_counter()
        await self.http.data_request("/", image=True)
        end = time.perf_counter()
        return (end - start)

    async def close(self) -> None:
        """|coro|
        Shuts down the asyncdagpi Client
        """
        await self.http.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asyncdagpi']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

setup_kwargs = {
    'name': 'asyncdagpi',
    'version': '4.1.5',
    'description': 'An Asynchronos API wrapper for https://dagpi.xyz',
    'long_description': '# asyncdagpi\n\n[![Build Status](https://travis-ci.com/Daggy1234/asyncdagpi.svg?branch=master)](https://travis-ci.com/Daggy1234/asyncdagpi) [![License](https://img.shields.io/github/license/daggy1234/asyncdagpi)](https://mit-license.org/) ![version](https://img.shields.io/pypi/v/asyncdagpi) [![python](https://img.shields.io/pypi/pyversions/asyncdagpi)](https://pypi.org/p/asyncdagpi) [![Documentation Status](https://readthedocs.org/projects/asyncdagpi/badge/?version=latest)](https://asyncdagpi.readthedocs.io/en/latest/?badge=latest) [![Codecov](https://img.shields.io/codecov/c/github/daggy1234/asyncdagpi?logo=codecov)](https://codecov.io/gh/daggy1234/asyncdagpi) [![Codacy Badge](https://app.codacy.com/project/badge/Grade/ad36f1ea6211444792e84f32a14326dd)](https://www.codacy.com/gh/Daggy1234/asyncdagpi/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=Daggy1234/asyncdagpi&amp;utm_campaign=Badge_Grade)\n\nPowerful Asynchronous Wrapper for dagpi [dagpi.xyz](https://dagpi.xyz)\n\n## Installation\n\n```shell script\npip install asyncdagpi\n```\n\n## Data API\n\nSome endpoints like WTP, PickupLine and Logo will return Objects while Waifu will return a Dictionary. Everything else will return a string.\n\n```python\nfrom asyncdagpi import Client\ndagpi = Client("dagpi token")\n# For WTP Object\nwtp = await dagpi.wtp()\n#For Roast\nroast = await dagpi.roast()\n```\n\n## Image Manipulation\n\nAll Image endpoints return an Image object. This has many properties that can be useful for developers. For Basic implementations are displayed\n\n### Discord.py\n\n```python\nfrom discord.ext import commands\nimport discord\nfrom asyncdagpi import Client, ImageFeatures\n\nbot = commands.Bot(command_prefix="!")\ndagpi = Client("dagpi token")\n\n@bot.command()\nasync def pixel(ctx, member: discord.Member):\n    url = str(member.display_avatar.with_format("png").with_size(1024))\n    img = await dagpi.image_process(ImageFeatures.pixel(), url)\n    file = discord.File(fp=img.image,filename=f"pixel.{img.format}")\n    await ctx.send(file=file)\n\n```\n\n### Pycord\n\n```python\nfrom discord.ext import commands\nimport discord\nfrom asyncdagpi import Client, ImageFeatures\n\nbot = discord.Bot()\ndagpi = Client("dagpi token")\n\n@bot.slash_command(guild_ids=[...])\nasync def pixel(\n    ctx, member: discord.Member = None\n):  # Passing a default value makes the argument optional\n    user = member or ctx.author\n    url = str(member.display_avatar.with_format("png").with_size(1024))\n    img = await dagpi.image_process(ImageFeatures.pixel(), url)\n    file = discord.File(fp=img.image,filename=f"pixel.{img.format}")\n\n\n```\n\n### Writing To File\n\n```python\nfrom asyncdagpi import Client, ImageFeatures\ndagpi = Client("dagpi token")\nimg = await dagpi.image_process(ImageFeatures.pixel(), "https://dagbot-is.the-be.st/logo.png")\n#it will auto chose the right format and write to current directory\nimg.write("pixel")\n#will create pixel.png in this case\n```\n\n### Python Pillow\n\n```python\nfrom asyncdagpi import ImageFeatures, Client\nfrom PIL import Image\n\ndagpi = Client("dagpi token")\nimg = await dagpi.image_process(ImageFeatures.pixel(), "https://dagbot-is.the-be.st/logo.png")\nim = Image.open(img.image)\n```\n\n### Kwargs Example\n\n```python\nfrom asyncdagpi import ImageFeatures, Client\n\ndagpi = Client("dagpi token")\nimg = await dagpi.image_process(ImageFeatures.tweet(), "https://dagbot-is.the-be.st/logo.png", text="This is asyncdagpi tweeting live from dagpi.xyz!", username="Asyncdagpi")\n```\n\n### For More Thorough Examples and Feature list read the documentation\n\n[Docs](https://asyncdagpi.rtfd.io)\n\n[Dagpi Docs](https://dagpi.docs.apiary.io)\n',
    'author': 'Daggy1234',
    'author_email': 'daggy@daggy.tech',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Daggy1234/asyncdagpi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
from typing import Dict, Any, List, Optional, TypeVar, Type
from datetime import datetime

from multidict import CIMultiDictProxy


RL = TypeVar('RL', bound='Ratelimits')


def maybe_convert(data: CIMultiDictProxy[str], key: str) -> Optional[int]:
    if val := data.get(key):
        try:
            return int(val)
        except BaseException:
            return None
    return None


class Ratelimits:

    def __init__(self, limit: Optional[int], remaining: Optional[int], reset: Optional[int]) -> None:
        self.limit: Optional[int] = limit
        self.remaining: Optional[int] = remaining
        self.reset: Optional[datetime] = datetime.fromtimestamp(reset) if reset else None

    @classmethod
    def from_dict(cls: Type[RL], data: CIMultiDictProxy[str]) -> RL:
        return cls(
            maybe_convert(
                data,
                'x-ratelimit-limit'),
            maybe_convert(
                data,
                'x-ratelimit-remaining'),
            maybe_convert(
                data,
                'x-ratelimit-reset'))

    def __repr__(self) -> str:
        return f"<asyncdagpi.Ratelimits limit={self.limit} remaining={self.remaining} reset={self.reset}>"


class BaseDagpiObject:

    """
    Base AsyncDagpi object

        .. container:: Attributes

            **Attributes**

            dict: :class:`Dict`
    """

    def __init__(self, dictionary: Dict[str, Any]):
        """
        Initialise a BaseDagpiObject
        """
        self.dict: Dict[str, Any] = dictionary

    def __str__(self) -> str:
        """
        Returns a string of the data
        :returns :class:`str`
        """
        return json.dumps(self.dict)

    def raw_data(self) -> Dict[str, Any]:
        """
        Dictionary with Raw Data
        :returns :class:`Dict`
        """
        return self.dict


class Logo(BaseDagpiObject):

    """
        A Dagpi Logo Object. A subclass of asyncdagpi.BaseDagpiObject

        .. container:: Attributes

            **Attributes**

            dict: :class:`Dict`
                Dictionary with raw data (same returned by raw_data)
            question: :class:`str`
                String containing url of question image
            answer: :class:`str`
                String containing url of answer image
            brand: :class:`str`
                String containing name of brand
            clue: :class:`str`
                String containing hint to use for question
            easy: :class:`boolean`
                Boolean that shows wether question was easy
            hint: :class:`str`
                String containing hint
            wiki_url: :class:`str`
                String containing Wikipedia URL for brand
        """

    def __init__(self, data: Dict[str, Any]):
        """
        Initialise a logo
        """
        super(Logo, self).__init__(data)
        self.dict: Dict[str, Any] = data
        self.question: str = data["question"]
        self.answer: str = data["answer"]
        self.brand: str = data["brand"]
        self.clue: Optional[str] = data.get("clue")
        self.easy: bool = data["easy"]
        self.hint: str = data["hint"]
        self.wiki_url: str = data["wiki_url"]


class PickupLine(BaseDagpiObject):

    """
    A Dagpi PickupLine

    .. container:: Attributes

        **Attributes**

        dict: :class:`Dict`
            Dictionary with raw data (same returned by raw_data)
        category: :class:`str`
            String containing Type of Pickup Line
        line: :class:`str`
            String Containing Pickup Line
    """

    def __init__(self, data: Dict[str, str]):
        super(PickupLine, self).__init__(data)
        self.dict: Dict[str, str] = data
        self.category: str = data["category"]
        self.line: str = data["joke"]


class WTP(BaseDagpiObject):

    """
           A Dagpi PickupLine. A subclass of AsyncDagpi object

           .. container:: Attributes

               **Attributes**

               dict: :class:`Dict`
                   Dictionary with raw data (same returned by raw_data)
               abilities: :class:`List[str]`
                   A list containing the abilities the pokemon has
               ascii: :class:`str`
                   String containing ascii data of pokemons image
               height: :class:`float`
                   Float with height of pokemon
               weight: :class:`float`
                   Float with weight of pokemon
               id: :class:`int`
                   Integer with national dex id of pokemon
               link: :class:`str`
                   String containing pokemondb link for pokemon
               name: :class:`str`
                   String containing name of pokemon
               answer: :class:`str`
                   String containing url for answer image
               question: :class:`str`
                   String containing url for question image
           """

    def __init__(self, data: Dict[str, Any]):
        super(WTP, self).__init__(data)
        mon: Dict[str, Any] = data["Data"]
        self.dict: Dict[str, Any] = data
        self.abilities: List[str] = mon["abilities"]
        self.ascii: str = mon["ascii"]
        self.height: float = mon["height"]
        self.id: int = int(mon["id"])
        self.link: str = mon["link"]
        self.name: str = mon["name"]
        self.type: List[str] = mon["Type"]
        self.weight: float = mon["weight"]
        self.question: str = data["question"]
        self.answer: str = data["answer"]


class Headline(BaseDagpiObject):

    """
    A Dagpi Headline

    .. container:: Attributes

        **Attributes**

        dict: :class:`Dict`
            Dictionary with raw data (same returned by raw_data)
        category: :class:`bool`
            Value stating whether headline is true or false
        headline: :class:`str`
            String Containing headline
    """

    def __init__(self, data: Dict[str, Any]):
        super(Headline, self).__init__(data)
        self.dict: Dict[str, Any] = data
        self.headline: str = data["text"]
        self.fake: bool = data["fake"]


class Captcha(BaseDagpiObject):
    """
    An asyncdagpi Captcha

    .. container:: Attributes

        **Attributes**

        dict: :class:`Dict`
            Dictionary with raw data (same returned by raw_data)
        image: :class:`str`
            url for image with solution
        answer: :class:`str`
            string with captcha value
    """

    def __init__(self, data: Dict[str, Any]):
        super(Captcha, self).__init__(data)
        self.dict: Dict[str, Any] = data
        self.image: str = data["image"]
        self.answer: str = data["answer"]


class Typeracer(BaseDagpiObject):
    """
    An asyncdagpi typeracer image

    .. container:: Attributes

        **Attributes**

        dict: :class:`Dict`
            Dictionary with raw data (same returned by raw_data)
        image: :class:`str`
            url for image with solution
        sentence: :class:`str`
            string with sentence
    """

    def __init__(self, data: Dict[str, Any]):
        super(Typeracer, self).__init__(data)
        self.dict: Dict[str, Any] = data
        self.image: str = data["image"]
        self.sentence: str = data["sentence"]
